{-|
Module      : Async.Worker
Description : Abstract async worker implementation using the 'Queue' typeclass
Copyright   : (c) Gargantext, 2024-Present
License     : AGPL
Maintainer  : gargantext@iscpif.fr
Stability   : experimental
Portability : POSIX
-}

{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE ScopedTypeVariables #-}


module Async.Worker
  ( run
  , sendJob
  , mkDefaultSendJob
  , sendJob'
  , SendJob(..) )
where


import Async.Worker.Broker
import Async.Worker.Types
import Control.Exception.Safe (catch, fromException, throwIO)
import Control.Monad (forever)
import System.Timeout qualified as Timeout


run :: (HasWorkerBroker b a) => State b a -> IO ()
run state@(State { .. }) = do
  createQueue broker queueName
  forever loop
  where
    loop :: IO ()
    loop = do
      -- TODO try...catch for main loop. This should catch exceptions
      -- but also job timeout events (we want to stick to the practice
      -- of keeping only one try...catch in the whole function)
      catch (do
                readMessageWaiting broker queueName >>= handleMessage state
            ) (\err ->
        case fromException err of
          Just jt@(JobTimeout {}) -> handleTimeoutError state jt
          Nothing -> case fromException err of
            Just je@(JobException {}) -> handleJobError state je
            _ -> undefined)

handleMessage :: (HasWorkerBroker b a) => State b a -> BrokerMessage b (Job a) -> IO ()
handleMessage state@(State { .. }) brokerMessage = do
  let msgId = messageId brokerMessage
  putStrLn $ formatStr state $ "received message with id = " <> show msgId
  let msg = getMessage brokerMessage
  let job = toA msg
  let mdata = metadata job
  let t = jobTimeout job
  mTimeout <- Timeout.timeout (t * microsecond) (performAction state brokerMessage)

  let archiveHandler = do
        case archiveStrategy mdata of
          ASDelete -> do
            -- putStrLn $ formatStr state $ "deleting completed job " <> show msgId <> " (strategy: " <> show archiveStrategy <> ")"
            deleteMessage broker queueName msgId
          ASArchive -> do
            -- putStrLn $ formatStr state $ "archiving completed job " <> show msgId <> " (strategy: " <> show archiveStrategy <> ")"
            archiveMessage broker queueName msgId

  
  case mTimeout of
    Just _ -> archiveHandler
    Nothing -> throwIO $ JobTimeout { jtBMessage = brokerMessage
                                    , jtTimeout = t }
  -- onMessageFetched broker queue msg

  -- (do
  --   let t = jobTimeout msg
  --   mTimeout <- Timeout.timeout t (handleJob broker msg)
  --   return ())

handleTimeoutError :: (HasWorkerBroker b a) => State b a -> JobTimeout b a -> IO ()
handleTimeoutError state@(State { .. }) jt@(JobTimeout { .. }) = do
  putStrLn $ formatStr state $ show jt
  let msgId = messageId jtBMessage
  let job = toA $ getMessage jtBMessage
  let mdata = metadata job
  case timeoutStrategy mdata of
    TSDelete -> deleteMessage broker queueName msgId
    TSArchive -> archiveMessage broker queueName msgId
    TSRepeat -> pure ()
    TSRepeatNElseArchive _n -> do
      -- TODO Implement 'readCt'
      pure ()
      -- OK so this can be repeated at most 'n' times, compare 'readCt' with 'n'
      -- if readCt > n then
      --   PGMQ.archiveMessage conn queue messageId
      -- else
      --   pure ()
    TSRepeatNElseDelete _n -> do
      -- TODO Implement 'readCt'
      pure ()
      -- OK so this can be repeated at most 'n' times, compare 'readCt' with 'n'
      -- if readCt > n then
      --   PGMQ.deleteMessage conn queue messageId
      -- else
      --   pure ()                      

handleJobError :: (HasWorkerBroker b a) => State b a -> JobException b a -> IO ()
handleJobError state@(State { .. }) je@(JobException {  .. }) = do
  putStrLn $ formatStr state $ show je
  let msgId = messageId jeBMessage
  let job = toA $ getMessage jeBMessage
  let mdata = metadata job
  case errorStrategy mdata of
    ESDelete -> deleteMessage broker queueName msgId
    ESArchive -> deleteMessage broker queueName msgId
    ESRepeat -> return ()


sendJob :: (HasWorkerBroker b a) => Broker b (Job a) -> Queue -> Job a -> IO ()
sendJob broker queueName job = do
  sendMessage broker queueName $ toMessage job

microsecond :: Int
microsecond = 10^(6 :: Int)


-- Job has quite a few metadata. Here are some utilities for
-- constructing them more easily

-- | Wraps parameters for the 'sendJob' function
data (HasWorkerBroker b a) => SendJob b a =
  SendJob { broker    :: Broker b (Job a)
          , queue     :: Queue
          , msg       :: a
          -- , delay     :: Delay
          , archStrat :: ArchiveStrategy
          , errStrat  :: ErrorStrategy
          , toStrat   :: TimeoutStrategy
          , timeout   :: Timeout }

-- | Create a 'SendJob' data with some defaults
mkDefaultSendJob :: HasWorkerBroker b a
                 => Broker b (Job a)
                 -> Queue
                 -> a
                 -> SendJob b a
mkDefaultSendJob broker queue msg =
  SendJob { broker
          , queue
          , msg
          -- , delay = 0
          -- | remove finished jobs
          , archStrat = ASDelete
          -- | archive errored jobs (for inspection later)
          , errStrat = ESArchive
          -- | repeat timed out jobs
          , toStrat = TSRepeat
          , timeout = 0 }
    
    
-- | Call 'sendJob' with 'SendJob b a' data
sendJob' :: (HasWorkerBroker b a) => SendJob b a -> IO ()
sendJob' (SendJob { .. }) = do
  let metadata = JobMetadata { archiveStrategy = archStrat
                             , errorStrategy = errStrat
                             , timeoutStrategy = toStrat
                             , timeout = timeout }
  let job = Job { job = msg, metadata }
  sendJob broker queue job
